import os
from io import BytesIO
from .movielens100k_data import MovieLens100kDataManager
import pandas as pd
from sklearn.model_selection import train_test_split

class MovieLens1MDataManager(MovieLens100kDataManager):
    DOWNLOAD_URL = 'http://files.grouplens.org/datasets/movielens/ml-1m.zip'
    DEFAULT_PATH = os.path.expanduser('~/.ml-1m.zip')
    
    def load_rating(self, random_state=114514, test_size=0.1):
        with BytesIO(self.zf.read('ml-1m/ratings.dat')) as ifs:
            import pandas as pd
            df = pd.read_csv(ifs, sep='\:\:', header=None, names=['user_id', 'movie_id', 'rating', 'timestamp'], engine='python')
            df['timestamp'] = pd.to_datetime(df.timestamp, unit='s')
            return train_test_split(df, random_state=random_state, test_size=test_size)


from zipfile import ZipFile
from io import BytesIO
import os
import urllib.request

import pandas as pd
from scipy import sparse as sps
from sklearn.model_selection import train_test_split


class MovieLens100kDataManager:
    DOWNLOAD_URL = 'http://files.grouplens.org/datasets/movielens/ml-100k.zip'
    DEFAULT_PATH = os.path.expanduser('~/.ml-100k.zip')
    
    @classmethod
    def _read_interaction(cls, byte_stream):
        with BytesIO(byte_stream) as ifs:
            data = pd.read_csv(ifs, sep='\t', header=None, names=['user_id', 'movie_id', 'rating', 'timestamp'])
            data['timestamp'] = pd.to_datetime(data['timestamp'], unit='s')
            return data

    def __init__(self, zippath=None):
        if zippath is None:
            zippath = self.DEFAULT_PATH
            if not os.path.exists(zippath):
                download = input('Could not find {}.\nCan I download and save it there?[y/N]'.format(zippath))
                if download.lower() == 'y':
                    print('start download...')
                    urllib.request.urlretrieve(self.DOWNLOAD_URL, zippath)
                    print('complete')
                else:
                    zippath = None
                    
        if zippath is not None:
            self.zf = ZipFile(zippath)
        else:
            self.zf = None
            
    def load_rating(self, random_state=114514, fold=None):
        if fold is None:
            df_all = self._read_interaction(self.zf.read('ml-100k/u.data'))
            df_train, df_test = train_test_split(df_all, random_state=random_state)
        else:
            assert fold >=1 and fold <=5
            train_path = 'ml-100k/u{}.base'.format(fold)
            test_path = 'ml-100k/u{}.test'.format(fold)
            df_train = self._read_interaction(self.zf.read(train_path))
            df_test = self._read_interaction(self.zf.read(test_path))
        return df_train, df_test
    
    def load_userinfo(self):
        user_info_bytes = self.zf.read('ml-100k/u.user')
        with BytesIO(user_info_bytes) as ifs:
            return pd.read_csv(ifs, sep='|', header=None, names=['user_id', 'age', 'gender', 'occupation', 'zipcode'])
    
    def load_movieinfo(self):
        MOVIE_COLUMNS = ['movie_id', 'title', 'release_date', 'unk', 'url']
        with BytesIO(self.zf.read('ml-100k/u.genre')) as ifs:
            genres = pd.read_csv(ifs, sep='|', header=None)[0]
        with BytesIO(self.zf.read('ml-100k/u.item')) as ifs:
            df_mov = pd.read_csv(
                ifs, sep='|', encoding='latin-1',
                header=None,
            )
            df_mov.columns = (
                MOVIE_COLUMNS + list(genres) 
            )
        df_mov['release_date'] = pd.to_datetime(df_mov.release_date)
        return df_mov, list(genres)


from collections import OrderedDict
import numpy as np
from scipy import (special, sparse as sps)
from tqdm import tqdm
from . import _myfm as core


def std_cdf(x):
    return (1 + special.erf(x * np.sqrt(.5))) / 2


def check_data_consistency(X, X_rel):
    shape = None
    if X_rel:
        shape_rel_all = {
            rel.mapper_size for rel in X_rel
        }
        if len(shape_rel_all) > 1:
            raise ValueError('Inconsistent case size for X_rel.')
        shape = list(shape_rel_all)[0]

    if X is None:
        if not X_rel:
            raise ValueError('At lease X or X_rel must be provided.')
    else:
        if shape is not None:
            if X.shape[0] != shape:
                raise ValueError('X and X_rel have different shape.')
        shape = X.shape[0]
    return shape


REAL = np.float64


class MyFMRegressor(object):

    def __init__(
        self, rank,
        init_stdev=0.1, random_seed=42,
        alpha_0=1.0, beta_0=1.0, gamma_0=1.0, mu_0=0.0, reg_0=1.0,
    ):
        """Factorization machine with Gibbs sampling.

        Parameters
        ----------
        rank : int
            The number of factors.

        init_stdev : float, optional (defalult = 0.1)
            The standard deviation for initialization.
            The factorization machine weights are randomely sampled from
            `Normal(0, init_stdev)`.

        random_seed : integer, optional (default = 0.1)
            The random seed used inside the whole learning process.

        alpha_0 : float, optional (default = 1.0)
            The half of alpha parameter for the gamma-distribution
            prior for alpha, lambda_w and lambda_V.
            Together with beta_0, the priors for these parameters are
            alpha, lambda_w, lambda_v ~ Gamma(alpha_0 / 2, beta_0 / 2)

        beta_0 : float, optioal (default = 1.0)
            See the explanation for alpha_0 .

        gamma_0: float optional (default = 1.0)
            Inverse variance of the prior for mu_w, mu_v.
            Together with mu_0, the priors for these parameters are
            mu_w, mu_v ~ Normal(mu_0, 1 / gamma_0)

        mu_0:
            See the explanation for gamma_0.

        reg_0:
            Inverse variance of tthe prior for w0.
            w0 ~ Normal(0, 1 / reg_0)
        """
        self.rank = rank

        self.init_stdev = init_stdev
        self.random_seed = random_seed

        self.alpha_0 = alpha_0
        self.beta_0 = beta_0
        self.gamma_0 = gamma_0
        self.mu_0 = mu_0

        self.reg_0 = reg_0

        self.predictor_ = None
        self.hypers_ = []

        self.n_groups_ = None

        
    def __str__(self):
        return "{class_name}(init_stdev={init_stdev}, alpha_0={alpha_0}, beta_0={beta_0}, gamma_0={gamma_0}, mu_0={mu_0}, reg_0={reg_0})".format(
            class_name=self.__class__.__name__,
            init_stdev=self.init_stdev,
            alpha_0=self.alpha_0, beta_0=self.beta_0,
            gamma_0=self.gamma_0, mu_0=self.mu_0,
            reg_0=self.reg_0
        )

    def fit(self, X, y, X_rel=[],
            X_test=None, y_test=None, X_rel_test=None,
            n_iter=100, n_kept_samples=None, grouping=None, callback=None, config_builder=None):
        """Performs Gibbs sampling to fit the data.
        Parameters
        ----------
        X : 2D array-like.
            Input variable.

        y : 1D array-like.
            Target variable.

        X_rel: list of RelationBlock, optional (defalult=[])
               Relation blocks which supplement X.

        n_iter : int, optional (defalult = 100)
            Iterations to perform.

        n_kept_samples: int, optional (default = None)
            The number of samples to store.
            If `None`, the value is set to `n_iter` - 5.

        grouping: Integer array, optional (default = None)
            If not `None`, this specifies which column of X belongs to which group.
            That is, if grouping[i] is g, then, w_i and V_{i, r}
            will be distributed according to
            Normal(mu_w[g], lambda_w[g]) and Normal(mu_V[g, r], lambda_V[g,r]),
            respectively.
            If `None`, all the columns of X are assumed to belong to a single group 0.

        callback: function(int, fm, hyper) -> bool, optional(default = None)
            Called at the every end of each iteration.
        """

        if config_builder is None:
            config_builder = core.ConfigBuilder()
        train_size = check_data_consistency(X, X_rel)
        if X is None:
            X = sps.csr_matrix((train_size, 0), dtype=np.float64)

        assert X.shape[0] == y.shape[0]
        dim_all = X.shape[1] + sum([rel.feature_size for rel in X_rel])

        if n_kept_samples is None:
            n_kept_samples = n_iter - 10
        else:
            assert n_iter >= n_kept_samples


        for key in ['alpha_0', 'beta_0', 'gamma_0', 'mu_0', 'reg_0']:
            value = getattr(self, key)
            getattr(config_builder, "set_{}".format(key))(value)
        if grouping is None:
            self.n_groups_ = 1
            config_builder.set_identical_groups(dim_all)
        else:
            assert dim_all == len(grouping)
            self.n_groups_ = np.unique(grouping).shape[0]
            config_builder.set_group_index(grouping)

        pbar = None
        if (X_test is not None or X_rel_test):
            if y_test is None:
                raise RuntimeError(
                    "Must specify both (X_test or X_rel_test) and y_test.")
            test_size = check_data_consistency(X_test, X_rel_test)
            assert test_size == y_test.shape[0]
            if X_test is None:
                X_test = sps.csr_matrix((test_size, 0), dtype=np.float64)
            do_test = True
        elif y_test is not None:
            raise RuntimeError(
                "Must specify both (X_test or X_rel_test) and y_test.")
        else:
            do_test = False

        config_builder.set_n_iter(n_iter).set_n_kept_samples(n_kept_samples)

        X = sps.csr_matrix(X)
        if X.dtype != np.float64:
            X.data = X.data.astype(np.float64)
        y = self.process_y(y)
        self.set_tasktype(config_builder)

        config = config_builder.build()

        if callback is None:
            pbar = tqdm(total=n_iter)

            def callback(i, fm, hyper):
                pbar.update(1)
                if i % 5:
                    return False

                log_str = self.status_report(fm, hyper)

                if do_test:
                    pred_this = self.process_score(
                        fm.predict_score(X_test, X_rel_test))
                    val_results = self.measure_score(pred_this, y_test)
                    for key, metric in val_results.items():
                        log_str += " {}_this: {:.2f}".format(key, metric)

                pbar.set_description(log_str)
                return False

        try:
            self.predictor_, self.hypers_ = \
                core.create_train_fm(self.rank, self.init_stdev, X, X_rel,
                                     y, self.random_seed, config, callback)
            return self
        finally:
            if pbar is not None:
                pbar.close()

    def set_tasktype(self, config_builder):
        config_builder.set_task_type(core.TaskType.REGRESSION)

    def predict(self, X, X_rel=[], n_workers=None):
        shape = check_data_consistency(X, X_rel)
        if X is None:
            X = sps.csr_matrix((shape, 0), dtype=np.float64)
        if n_workers is not None:
            return self.predictor_.predict_parallel(X, X_rel, n_workers)
        else:
            return self.predictor_.predict(X, X_rel)

    @classmethod
    def status_report(cls, fm, hyper):
        log_str = "alpha = {:.2f} ".format(hyper.alpha)
        log_str += "w0 = {:.2f} ".format(fm.w0)
        return log_str

    @classmethod
    def process_score(cls, y):
        return y

    @classmethod
    def process_y(cls, y):
        return y

    @classmethod
    def measure_score(cls, prediction, y):
        result = OrderedDict()
        result['rmse'] = ((y - prediction) ** 2).mean() ** 0.5
        result['mae'] = np.abs(y - prediction).mean()
        return result

    def get_hyper_trace(self, dataframe=True):
        columns = (
            ['alpha'] +
            ['mu_w[{}]'.format(g) for g in range(self.n_groups_)] +
            ['lambda_w[{}]'.format(g) for g in range(self.n_groups_)] +
            ['mu_V[{},{}]'.format(g, r) for g in range(self.n_groups_) for r in range(self.rank)] +
            ['lambda_V[{},{}]'.format(g, r) for g in range(
                self.n_groups_) for r in range(self.rank)]
        )

        res = []
        for hyper in self.hypers_:
            res.append(np.concatenate([
                [hyper.alpha], hyper.mu_w, hyper.lambda_w,
                hyper.mu_V.ravel(), hyper.lambda_V.ravel()
            ]))
        res = np.vstack(res)
        if dataframe:
            import pandas as pd
            res = pd.DataFrame(res)
            res.columns = columns
            return res
        else:
            return [
                {key: sample[i] for i, key in enumerate(columns)}
                for sample in res
            ]


class MyFMClassifier(MyFMRegressor):
    def set_tasktype(self, config_builder):
        config_builder.set_task_type(core.TaskType.CLASSIFICATION)

    @classmethod
    def process_score(cls, score):
        return std_cdf(score)

    @classmethod
    def process_y(cls, y):
        return y.astype(np.float64) * 2 - 1

    @classmethod
    def measure_score(cls, prediction, y):
        result = OrderedDict()
        lp = np.log(prediction + 1e-15)
        l1mp = np.log(1 - prediction + 1e-15)
        gt = y > 0
        result['ll'] = - lp.dot(gt) - l1mp.dot(~gt)
        result['accuracy'] = np.mean((prediction >= 0.5) == gt)
        return result

    @classmethod
    def status_report(cls, fm, hyper):
        log_str = "w0 = {:.2f} ".format(fm.w0)
        return log_str

    def predict(self, *args, **kwargs):
        return (self.predict_proba(*args, **kwargs)) > 0.5

    def predict_proba(self, *args, **kwargs):
        return super().predict(*args, **kwargs)


class MyFMOrderedProbit(MyFMRegressor):
    def set_tasktype(self, config_builder):
        config_builder.set_task_type(core.TaskType.ORDERED)

    def fit(self, X, y, X_rel=[],
            n_iter=100, n_kept_samples=None, grouping=None, callback=None,
            cutpoint_group_configs=None
    ):
        config_builder = core.ConfigBuilder()
        y = np.asarray(y)
        if cutpoint_group_configs is None:
            n_class = y.max() + 1
            cutpoint_group_configs = [
                (int(n_class), np.arange(y.shape[0], dtype=np.int64))
            ]
        self.n_cutpoint_groups = len(cutpoint_group_configs)
        config_builder.set_cutpoint_groups(
            cutpoint_group_configs
        )
        return super().fit(
            X, y, X_rel=X_rel,
            n_iter=n_iter, n_kept_samples=n_kept_samples,
            grouping=grouping, callback=callback,
            config_builder=config_builder
        )

    @classmethod
    def process_score(cls, score):
        return (1 + special.erf(score * np.sqrt(.5))) / 2

    @classmethod
    def process_y(cls, y):
        y_as_float = y.astype(np.float64)
        assert y.min() >= 0
        return y_as_float

    @classmethod
    def measure_score(cls, prediction, y):
        raise NotImplementedError('not implemented')
        result = OrderedDict()
        lp = np.log(prediction + 1e-15)
        l1mp = np.log(1 - prediction + 1e-15)
        gt = y > 0
        result['ll'] = - lp.dot(gt) - l1mp.dot(~gt)
        result['accuracy'] = np.mean((prediction >= 0.5) == gt)
        return result

    @classmethod
    def status_report(cls, fm, hyper):
        
        log_str = "w0= {:2f}".format(fm.w0)
        if len(fm.cutpoints) == 1:
            log_str += ", cutpoint = {} ".format(
                ["{:.3f}".format(c) for c in list(fm.cutpoints[0])]
            )
        return log_str


    def predict_proba(self, X, rels=[], cutpoint_index=None, **kwargs):
        if cutpoint_index is None:
            if self.n_cutpoint_groups == 1:
                cutpoint_index = 0
        else:
            raise ValueError('specify the cutpoint index')
        X = sps.csr_matrix(X)
        if X.dtype != np.float64:
            X.data = X.data.astype(np.float64)
        p = 0

        sample_offset = len(self.hypers_) - len(self.predictor_.samples)
        for sample_index, sample in enumerate(self.predictor_.samples):
            alpha = self.hypers_[sample_index + sample_offset].alpha
            score = sample.predict_score(X, rels)
            score = std_cdf(
                np.sqrt(alpha) * (sample.cutpoints[cutpoint_index][np.newaxis, :] - score[:, np.newaxis])
            )
            score = np.hstack([
                np.zeros((score.shape[0], 1), dtype=score.dtype),
                score,
                np.ones((score.shape[0], 1), dtype=score.dtype)
            ])
            p += (score[:, 1:] - score[:, :-1])
        return p / len(self.predictor_.samples)

    def predict(self, *args, **kwargs):
        return self.predict_proba(*args, **kwargs).argmax(axis=1)



from collections import OrderedDict
from abc import ABC, abstractmethod
from scipy import special
from tqdm import tqdm
import pandas as pd
import numpy as np

def std_cdf(x):
    return (1 + special.erf(x * np.sqrt(.5))) / 2


class LibFMLikeCallbackBase(ABC):
    def __init__(self, n_iter, X_test, X_rel_test,
                 y_test, trace_path=None):
        """Provides a LibFM-like callback after each iteration.
        This will be helpful when we cannot afford enough memory to store
        all posterior samples."""

        self.n_iter = n_iter
        self.X_test = X_test
        self.X_rel_test = X_rel_test
        self.y_test = y_test
        self.result_trace = []
        self.pbar = None
        self.trace_path = trace_path
        self.n_samples = 0

    def __enter__(self):
        self.pbar = tqdm(total=self.n_iter)

    def __exit__(self, *args):
        self.pbar.close()
        self.pbar = None
    
    @abstractmethod
    def _measure_score(self, i, fm, hyper):
        return None, None

    def __call__(self, i, fm, hyper):
        description,  trace_result = self._measure_score(i, fm, hyper)

        if self.pbar is not None:
            self.pbar.set_description(
                description
            )
        self.result_trace.append(trace_result)
        if self.trace_path is not None:
            df = pd.DataFrame(self.result_trace)
            df.to_csv(self.trace_path, index=False)

        if self.pbar is not None:
            self.pbar.update(1)

        return False

class RegressionCallback(LibFMLikeCallbackBase):
    def __init__(self, n_iter,
                 X_test, y_test, X_rel_test=None,
                 clip_min=None, clip_max=None, trace_path=None):
        super(RegressionCallback, self).__init__(
            n_iter, X_test, X_rel_test, y_test, trace_path=trace_path
        )
        self.predictions =  np.zeros((X_test.shape[0], ), dtype=np.float64)
        self.prediction_all_but_5 = np.zeros((X_test.shape[0], ), dtype=np.float64)
        self.clip_min = clip_min
        self.clip_max = clip_max

    def clip_value(self, arr):
        if self.clip_min is not None:
            arr[arr <= self.clip_min] = self.clip_min
        if self.clip_max is not None:
            arr[arr >= self.clip_max] = self.clip_max

    def _measure_score(self, i, fm, hyper):
        score = fm.predict_score(self.X_test, self.X_rel_test)
        self.predictions += score
        self.n_samples +=1
        prediction_mean = (self.predictions / self.n_samples)
        self.clip_value(prediction_mean)
        if i >= 5:
            self.prediction_all_but_5 += score
            prediction_mean_all_but_5 = self.prediction_all_but_5 / ( i + 1 - 5)
            self.clip_value(prediction_mean_all_but_5)
            rmse_all_but_5 = float(((self.y_test - prediction_mean_all_but_5) ** 2).mean() ** 0.5)
        else:
            rmse_all_but_5 = float('nan')

        rmse =  float(( (self.y_test - prediction_mean) ** 2 ).mean() ** 0.5)
        rmse_this = float(( (self.y_test - score) ** 2 ).mean() ** 0.5)
        description = \
            "alpha={0:.4f}, rmse_mean={1:.4f}, rmse_this={2:.4f}, rmse_all_but_5={3:.4f}".format(
                hyper.alpha, rmse, rmse_this, rmse_all_but_5
            )
        result = OrderedDict(
            [
                ('alpha', hyper.alpha), ('rmse', rmse),
                ('rmse_this', rmse_this), ('rmse_all_but_5', rmse_all_but_5)
            ]
        )
        return description, result

class ClassificationCallback(LibFMLikeCallbackBase):
    def __init__(self, n_iter,
                 X_test, y_test, X_rel_test=None, eps=1e-15,
                 trace_path="rmse_result.csv"):
        super(ClassificationCallback, self).__init__(
            n_iter, X_test, X_rel_test, y_test, trace_path=trace_path
        )
        self.predictions =  np.zeros((X_test.shape[0], ), dtype=np.float64)
        self.prediction_all_but_5 = np.zeros((X_test.shape[0], ), dtype=np.float64)
        self.eps = eps
    
    def clip_value(self, arr):
        if self.eps is not None:
            arr[arr <= self.eps] = self.eps
            arr[arr >= (1 - self.eps)] = 1 - self.eps
    
    def __log_loss(self, arr): 
        result = 0
        result += np.log(arr[self.y_test==1]).sum()
        result += np.log(1 - arr[self.y_test==0]).sum()
        return - result

    def __accuracy(self, arr): 
        return float((self.y_test == (arr >= 0.5)).mean())

    def _measure_score(self, i, fm, hyper):
        prob_this = fm.predict_proba(self.X_test, self.X_rel_test)
        self.predictions += prob_this
        self.n_samples +=1
        prediction_mean = (self.predictions / self.n_samples)
        self.clip_value(prediction_mean)
        if i >= 5:
            self.prediction_all_but_5 += prob_this
            prediction_mean_all_but_5 = self.prediction_all_but_5 / ( i + 1 - 5)
            self.clip_value(prediction_mean_all_but_5)
            ll_all_but_5 = self.__log_loss(prediction_mean_all_but_5)              
            accuracy_all_but_5 = self.__accuracy(prediction_mean_all_but_5)
        else:
            ll_all_but_5 = float('nan')
            accuracy_all_but_5 = float('nan')

        ll =  self.__log_loss(prediction_mean)
        accuracy = self.__accuracy(prediction_mean)
        ll_this = self.__log_loss(prob_this)
        accuracy_this = self.__accuracy(prob_this)
        description = \
            "ll_mean={0:.4f}, ll_this={1:.4f}, ll_all_but_5={2:.4f}".format(
                ll, ll_this, ll_all_but_5
            )
        result = OrderedDict(
            [
                ('log_loss', ll),
                ('log_loss_this', ll_this),
                ('log_loss_all_but_5', ll_all_but_5),
                ('accuracy', accuracy),
                ('accuracy_this', accuracy_this),
                ('accuracy_all_but_5', accuracy_all_but_5)
            ]
        )
        return description, result

class OrderedProbitCallback(LibFMLikeCallbackBase):
    def __init__(self, n_iter,
                 X_test, y_test, n_class, X_rel_test=None, 
                 eps=1e-15,
                 trace_path=None):
        super(OrderedProbitCallback, self).__init__(
            n_iter, X_test, X_rel_test, y_test, trace_path=trace_path
        )
        self.predictions =  np.zeros((X_test.shape[0], n_class), dtype=np.float64)
        self.prediction_all_but_5 = np.zeros((X_test.shape[0], n_class ), dtype=np.float64)
        self.n_class = n_class
        self.eps = eps
        self.y_test = self.y_test.astype(np.int32)
        assert (self.y_test.min() >= 0) and (self.y_test.max() <= (self.n_class - 1))
    
    
    def __log_loss(self, arr): 
        ps = arr[np.arange(self.y_test.shape[0]), self.y_test].copy()
        ps[ps <= self.eps] = self.eps
        return - float(np.log(ps).sum())

    def __accuracy(self, arr): 
        return float((self.y_test == (arr.argmax(axis=1))).mean())
    
    def __rmse(self, arr):
        return float((
            (self.y_test - arr.dot(np.arange(self.n_class))) ** 2
        ).mean()) ** 0.5

    def _measure_score(self, i, fm, hyper):
        score = fm.predict_score(self.X_test, self.X_rel_test)
        score = std_cdf(
            fm.cutpoints[0][np.newaxis, :] - score[:, np.newaxis]
        )
        score = np.hstack([
            np.zeros((score.shape[0], 1), dtype=score.dtype),
            score,
            np.ones((score.shape[0], 1), dtype=score.dtype)
        ])
        prob_this = score[:, 1:] - score[:, :-1]
        self.predictions += prob_this
        self.n_samples +=1
        prediction_mean = (self.predictions / self.n_samples)
        if i >= 5:
            self.prediction_all_but_5 += prob_this
            prediction_mean_all_but_5 = self.prediction_all_but_5 / ( i + 1 - 5)
            ll_all_but_5 = self.__log_loss(prediction_mean_all_but_5)              
            accuracy_all_but_5 = self.__accuracy(prediction_mean_all_but_5)
            rmse_all_but_5 = self.__rmse(prediction_mean_all_but_5)
        else:
            ll_all_but_5 = float('nan')
            accuracy_all_but_5 = float('nan')
            rmse_all_but_5 = float('nan')

        ll =  self.__log_loss(prediction_mean)
        accuracy = self.__accuracy(prediction_mean)
        rmse = self.__rmse(prediction_mean)
        ll_this = self.__log_loss(prob_this)
        accuracy_this = self.__accuracy(prob_this)
        rmse_this = self.__rmse(prob_this)
        description = \
            "ll_mean={0:.4f}, ll_this={1:.4f}, ll_all_but_5={2:.4f}".format(
                ll, ll_this, ll_all_but_5
            )
        result = OrderedDict(
            [
                ('log_loss', ll),
                ('log_loss_this', ll_this),
                ('log_loss_all_but_5', ll_all_but_5),
                ('accuracy', accuracy),
                ('accuracy_this', accuracy_this),
                ('accuracy_all_but_5', accuracy_all_but_5),
                ('rmse', rmse),
                ('rmse_this', rmse_this),
                ('rmse_all_but_5', rmse_all_but_5)
            ]
        )
        return description, result


# Taken from
# https://github.com/wichert/pybind11-example/blob/master/setup.py  
# and modified.

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
import sys
import setuptools
from distutils.command.clean import clean as Clean
import os

__version__ = '0.2.1.1'

install_requires = ['pybind11>=2.5.0', 'numpy>=1.11', 'scipy>=1.0', 'tqdm>=4']

eigen_include_dir = os.environ.get('EIGEN3_INCLUDE_DIR', None)
if eigen_include_dir is None:
    install_requires.append('requests')


class get_eigen_include(object):
    EIGEN3_URL = 'https://gitlab.com/libeigen/eigen/-/archive/3.3.7/eigen-3.3.7.zip'
    EIGEN3_DIRNAME = 'eigen-3.3.7'

    def __str__(self):

        if eigen_include_dir is not None:
            return eigen_include_dir

        basedir = os.path.dirname(__file__)
        target_dir = os.path.join(basedir, self.EIGEN3_DIRNAME)
        if os.path.exists(target_dir):
            return target_dir

        download_target_dir = os.path.join(basedir, 'eigen3.zip')
        import requests
        import zipfile
        print('Start downloading Eigen library from {}.'.format(self.EIGEN3_DIRNAME))
        response = requests.get(self.EIGEN3_URL, stream=True)
        with open(download_target_dir, 'wb') as ofs:
            for chunk in response.iter_content(chunk_size=1024):
                ofs.write(chunk)
        print('Downloaded Eigen into {}.'.format(download_target_dir))

        with zipfile.ZipFile(download_target_dir) as ifs:
            ifs.extractall()

        return target_dir


class get_pybind_include(object):
    """Helper class to determine the pybind11 include path
    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)

headers = [
    'include/myfm/definitions.hpp',
    'include/myfm/util.hpp',
    'include/myfm/FM.hpp',
    'include/myfm/HyperParams.hpp',
    'include/myfm/predictor.hpp',
    'include/myfm/FMTrainer.hpp',
    'include/myfm/FMLearningConfig.hpp',
    'include/myfm/OProbitSampler.hpp',
    'include/Faddeeva/Faddeeva.hh',
    'src/declare_module.hpp'
]


ext_modules = [
    Extension(
        'myfm._myfm',
        ['src/bind.cpp', 'src/Faddeeva.cc'],
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            get_eigen_include(),
            "include"
        ],
        language='c++'
    ),
]


# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14/17] compiler flag.
    The newer version is prefered over c++11 (when it is available).
    """
    flags = ['-std=c++11']

    for flag in flags:
        if has_flag(compiler, flag):
            return flag

    raise RuntimeError('Unsupported compiler -- at least C++11 support '
                       'is needed!')


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': ['-O3'],
    }
    l_opts = {
        'msvc': [],
        'unix': [],
    }

    if sys.platform == 'darwin':
        darwin_opts = ['-stdlib=libc++', '-mmacosx-version-min=10.7']
        c_opts['unix'] += darwin_opts
        l_opts['unix'] += darwin_opts

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        link_opts = self.l_opts.get(ct, [])
        if ct == 'unix':
            opts.append('-DVERSION_INFO="%s"' %
                        self.distribution.get_version())
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, '-fvisibility=hidden'):
                opts.append('-fvisibility=hidden')
        elif ct == 'msvc':
            opts.append('/DVERSION_INFO=\\"%s\\"' %
                        self.distribution.get_version())
        for ext in self.extensions:
            ext.extra_compile_args = opts
            ext.extra_link_args = link_opts
        build_ext.build_extensions(self)


setup(
    name='myfm',
    version=__version__,
    author='Tomoki Ohtsuki',
    url='https://github.com/tohtsky/myfm',
    author_email='tomoki.ohtsuki129@gmail.com',
    description='Yet another factorization machine',
    long_description='',
    ext_modules=ext_modules,
    install_requires=install_requires,
    setup_requires=install_requires,
    cmdclass={'build_ext': BuildExt},
    packages=['myfm', 'myfm.utils.benchmark_data', 'myfm.utils.callbacks'],
    zip_safe=False,
    headers=headers
)
